from fastapi import (
    APIRouter,
    Depends,
    Response,
    status,
    Request,
    Query,
    Path
)
from fastapi.responses import JSONResponse
from fastapi_keycloak_sso.initializer import RoutAccessInitializer
from fastapi_keycloak_sso.api.schemas import (
    LoginSchema,
    GiveTokenSchema,
    GroupSchema,
    PaginatedGroups,
    UserSchema,
    GroupCreateSchema,
    AssignRoleGroupManySchema,
    UserJoinGroupSchema
)
from fastapi_keycloak_sso.keycloak import KeyCloakConfidentialClient
from fastapi_keycloak_sso.auth import (
    authenticate,
    UserPayload
)
from fastapi_keycloak_sso.api.enums import DetailType
from fastapi_keycloak_sso.decorators import require_any_group


router = APIRouter(prefix='/keycloak')

# to routs access
access = {
    'group_read': RoutAccessInitializer.group_read_access,
    'group_find': RoutAccessInitializer.group_find_access,
    'group_create': RoutAccessInitializer.group_create_access,
    'group_delete': RoutAccessInitializer.group_delete_access,
    'users_read': RoutAccessInitializer.user_read_access,
    'roles_read': RoutAccessInitializer.role_read_access,
    'assign_role_group': RoutAccessInitializer.assign_role_group_access,
    'join_user_group': RoutAccessInitializer.join_user_group_access
}


@router.post('/token/', status_code=status.HTTP_200_OK, responses={
     200: {"description": "string"},
     400: {'description': {'detail':'string'}}
}, tags=['KeyCloak - Accounts'])
def give_access_token(body: GiveTokenSchema):
    """
    Create token for given user
    """
    keycloak = KeyCloakConfidentialClient()

    username = body.username
    password = body.password

    try:
        access_key = keycloak.send_request(
            request_type= keycloak.KeyCloakRequestTypeChoices.PASSWORD_ACCESS_TOKEN,
            request_type_choices= keycloak.KeyCloakRequestTypeChoices,
            request_method= keycloak.KeyCloakRequestMethodChoices.POST,
            panel_type= keycloak.KeyCloakPanelTypeChoices.USER,
            username=username,
            password=password
        )
    except keycloak.KeyCloakException as e:
        return JSONResponse({'detail': str(e)}, status_code=status.HTTP_400_BAD_REQUEST)
    return JSONResponse(access_key, status_code=status.HTTP_200_OK)


@router.post('/login/', status_code=status.HTTP_200_OK, responses={
    200: {'message':'string'},
    401: {
        "description": "Unauthorized",
        "content": {
            "application/json": {
                "example": {"error": "string"}
            }
        }}
},tags=['KeyCloak - Accounts'])
def login(body: LoginSchema):
    """
    Set received token from keycloak on request cookie
    """
    keycloak_klass = KeyCloakConfidentialClient()

    try:
        decoded_token = keycloak_klass.decode_token(body.token)
    except KeyCloakConfidentialClient.KeyCloakException as e:
        return Response({"error": str(e)}, status_code=status.HTTP_401_UNAUTHORIZED)

    response = JSONResponse({
        "message": "Login successful",
        "token": body.token,
        "user": decoded_token
    }, status_code=status.HTTP_200_OK)

    if keycloak_klass.save_token_method == keycloak_klass.KeyCloakSaveTokenMethodChoices.COOKIE:
        response = keycloak_klass.set_httponly_cookie('access_token', body.token, response)
        response = keycloak_klass.set_httponly_cookie('refresh_token', body.refresh_token, response)
        response = keycloak_klass.set_httponly_cookie('client_id', body.client_id, response)

    return response

@router.post('/token/refresh/', status_code=status.HTTP_200_OK, responses={
    200: {"description": "string"},
    401: {
        "description": "No refresh token or client ID / Token refresh failed",
        "content": {
            "application/json": {
                "example": {"detail": "string"}
            }
        }}
},tags=['KeyCloak - Accounts'])
def refresh_token(request: Request):
    """
    Refresh received token from keycloak
    """

    keycloak_klass = KeyCloakConfidentialClient()
    refresh_token = keycloak_klass.get_token(request, 'refresh_token')
    client_id = keycloak_klass.get_token(request, 'client_id')

    if not refresh_token or not client_id:
        return JSONResponse({"detail": "No refresh token Or Client ID"}, status_code=status.HTTP_401_UNAUTHORIZED)

    try:
        new_tokens = keycloak_klass.send_request(
            keycloak_klass.KeyCloakRequestTypeChoices.REFRESH_ACCESS_TOKEN,
            keycloak_klass.KeyCloakRequestTypeChoices,
            keycloak_klass.KeyCloakRequestMethodChoices.POST,
            keycloak_klass.KeyCloakPanelTypeChoices.USER,
            refresh_token=refresh_token,
            client_id=client_id,
        )
    except Exception as e:
        return JSONResponse({"detail": "string"}, status_code=status.HTTP_401_UNAUTHORIZED)

    response = JSONResponse({"detail": "Token refreshed"}, status_code=status.HTTP_200_OK)
    response.set_cookie("access_token", new_tokens["access_token"], httponly=True, secure=True)

    if keycloak_klass.save_token_method == keycloak_klass.KeyCloakSaveTokenMethodChoices.COOKIE:
        response = keycloak_klass.set_httponly_cookie('refresh_token', refresh_token, response)
        if "refresh_token" in new_tokens:
            response.set_cookie("refresh_token", new_tokens["refresh_token"], httponly=True, secure=True)
        return response

    return JSONResponse(new_tokens, status_code=status.HTTP_200_OK)


@router.post('/logout/', status_code=status.HTTP_200_OK, responses = {
            200: {
                'description': 'Logged out successfully',
                'content': {
                    'application/json': {
                        'schema': {
                            'type': 'object',
                            'properties': {
                                'detail': {'type': 'string', 'example': 'Logged out'}
                            }
                        }
                    }
                }
            },
            401: {
                'description': 'No refresh token or client ID',
                'content': {
                    'application/json': {
                        'schema': {
                            'type': 'object',
                            'properties': {
                                'detail': {'type': 'string'}
                            }
                        }
                    }
                  }
            }
},tags=['KeyCloak - Accounts'])
def logout(request: Request, user: UserPayload = Depends(authenticate)):
    """
    Logging out the user and deleting data from cookies
    """

    keycloak_klass = KeyCloakConfidentialClient()
    refresh_token = keycloak_klass.get_token(request, 'refresh_token')
    client_id = keycloak_klass.get_token(request, 'client_id')

    if not refresh_token or not client_id:
        return JSONResponse({"detail": "No refresh token Or Client ID"}, status_code=status.HTTP_401_UNAUTHORIZED)

    logout_res = keycloak_klass.send_request(
        keycloak_klass.KeyCloakRequestTypeChoices.LOGOUT,
        keycloak_klass.KeyCloakRequestTypeChoices,
        keycloak_klass.KeyCloakRequestMethodChoices.POST,
        keycloak_klass.KeyCloakPanelTypeChoices.USER,
        refresh_token=refresh_token,
        client_id=client_id,
    )

    if keycloak_klass.save_token_method == keycloak_klass.KeyCloakSaveTokenMethodChoices.COOKIE:
        response = JSONResponse({"detail": "Logged out"}, status_code=status.HTTP_200_OK)
        response.delete_cookie("access_token")
        response.delete_cookie("refresh_token")
        response.delete_cookie("client_id")
        return response

    return JSONResponse({"detail": "Logged out"}, status_code=status.HTTP_200_OK)


@router.get("/groups/", response_model=PaginatedGroups,tags=['KeyCloak - Admin'])
@require_any_group(*access['group_read'])
def group_read(
    pk: str = None,
    limit: int = Query(10, ge=1, le=100,
                       description='Maximum number of results to return (page size). Default is 10.'),
    offset: int = Query(0, ge=0,
                        description='Number of items to skip before starting to return results (starting index)'),
    user: UserPayload = Depends(authenticate)
):
    """
    Reading existing groups in the Realm system
    """

    keycloak_klass = KeyCloakConfidentialClient()

    try:
        response = keycloak_klass.send_request(
            keycloak_klass.KeyCloakRequestTypeChoices.GROUPS,
            keycloak_klass.KeyCloakRequestTypeChoices,
            keycloak_klass.KeyCloakRequestMethodChoices.GET,
            keycloak_klass.KeyCloakPanelTypeChoices.ADMIN,
            detail_pk=pk,
        )

        # Normalize empty or single result
        if response is None:
            response = []
        elif not isinstance(response, list):
            response = [response]

        # Convert raw dicts → Pydantic models
        groups = [GroupSchema.from_instance(item) for item in response]

        total = len(groups)

        # Apply pagination
        paginated = groups[offset : offset + limit]

        return PaginatedGroups(
            count=total,
            limit=limit,
            offset=offset,
            results=paginated
        )

    except keycloak_klass.KeyCloakNotFoundException:
        return JSONResponse({"detail": "Requested group not found"}, status_code=404)
    
@router.get('/profile/', response_model=UserSchema, status_code=status.HTTP_200_OK,
            tags=['KeyCloak - Admin'])
def user_profile(user: UserPayload = Depends(authenticate)):
    """
    Read profile data
    """
    return user.profile


def _find_group_exact(groups, search_name):
    """
    Recursively find the exact group (name matches exactly).
    Returns the full group object.
    """
    for group in groups:
        if group.get("name") == search_name:
            return group

        sub_groups = group.get("subGroups") or []
        if sub_groups:
            result = _find_group_exact(sub_groups, search_name)
            if result:
                return result

    return None


@router.get('/group/{group_name}/',tags=['KeyCloak - Admin'])
@require_any_group(*access['group_find'])
def find_group_detail_exact(
    group_name: str = Path(description='Enter group name'),
    detailing_type: DetailType = Query(
        default=DetailType.id,
        description="Type of detail to return: 'id' for group ID, 'full' for full group object"),
    user: UserPayload = Depends(authenticate)
):
    """
    Get exact group by name.
    """
    keycloak = KeyCloakConfidentialClient()

    # Extra params: search param helps Keycloak filter server-side
    extra_params = {'extra_query_params': {'search': group_name}}

    try:
        groups = keycloak.send_request(
            keycloak.KeyCloakRequestTypeChoices.GROUPS,
            keycloak.KeyCloakRequestTypeChoices,
            keycloak.KeyCloakRequestMethodChoices.GET,
            keycloak.KeyCloakPanelTypeChoices.ADMIN,
            **extra_params
        )

        group = _find_group_exact(groups, group_name)

        if not group:
            return JSONResponse(
                {"detail": "Group name not found"},
                status_code=status.HTTP_404_NOT_FOUND
            )

        if detailing_type == DetailType.id:
            return {"id": group.get("id")}

        return group

    except Exception as e:
        return JSONResponse(
            {"detail": "Error fetching groups", "error": str(e)},
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR
        )

@router.get('/users/',tags=['KeyCloak - Admin'])
@require_any_group(*access['users_read'])
def users_read(
        user: UserPayload = Depends(authenticate),
        pk: str = None,
):
    """
    Read user information.
    """
    keycloak_klass = KeyCloakConfidentialClient()

    try:
        response = keycloak_klass.send_request(
            keycloak_klass.KeyCloakRequestTypeChoices.USERS,
            keycloak_klass.KeyCloakRequestTypeChoices,
            keycloak_klass.KeyCloakRequestMethodChoices.GET,
            keycloak_klass.KeyCloakPanelTypeChoices.ADMIN,
            detail_pk=pk,
        )
    except keycloak_klass.KeyCloakNotFoundException as e:
        return JSONResponse(
            {"detail": "Requested user data was not found"},
            status_code=status.HTTP_404_NOT_FOUND)

    return JSONResponse(response, status_code=status.HTTP_200_OK)

@router.post('/groups/create/',tags=['KeyCloak - Admin'])
@require_any_group(*access['group_create'])
def group_create(
    body: GroupCreateSchema,
    group_parent_id: str = Query(default=None,description='Enter the group ID above of the branch.'),
    user: UserPayload = Depends(authenticate)
):
    """
    Create a group in a realm or create a group in a subdirectory of another group.
    """
    keycloak = KeyCloakConfidentialClient()

    group_name = body.name

    try:
        response = keycloak.send_request(
            keycloak.KeyCloakRequestTypeChoices.GROUPS,
            keycloak.KeyCloakRequestTypeChoices,
            keycloak.KeyCloakRequestMethodChoices.POST,
            keycloak.KeyCloakPanelTypeChoices.ADMIN,
            name=group_name,
            group_parent_id=group_parent_id,
        )

    except Exception as e:
        return JSONResponse({
            'detail': str(e)
        },
            status_code=status.HTTP_400_BAD_REQUEST
        )

    return JSONResponse(response, status_code=status.HTTP_200_OK)


@router.delete('/groups/{group_id}/',tags=['KeyCloak - Admin'])
@require_any_group(*access['group_delete'])
def group_delete(
        group_id: str,
        user: UserPayload = Depends(authenticate)
):
    """
    Delete group by id.
    """
    keycloak = KeyCloakConfidentialClient()

    try:
        response = keycloak.send_request(
            keycloak.KeyCloakRequestTypeChoices.GROUPS,
            keycloak.KeyCloakRequestTypeChoices,
            keycloak.KeyCloakRequestMethodChoices.DELETE,
            keycloak.KeyCloakPanelTypeChoices.ADMIN,
            group_id=group_id
        )

    except Exception as e:
        return JSONResponse({'detail':str(e)},
                            status_code=status.HTTP_400_BAD_REQUEST)
    return JSONResponse(response, status_code=status.HTTP_200_OK)

@router.get('/roles/',tags=['KeyCloak - Admin'])
@require_any_group(*access['roles_read'])
def role_read(
        user: UserPayload = Depends(authenticate),
        role_id: str = Query(default=None,description='Enter the role ID.'),
):
    """
    Reading client roles
    """
    keycloak = KeyCloakConfidentialClient()
    try:
        response = keycloak.send_request(
            keycloak.KeyCloakRequestTypeChoices.CLIENT_ROLES,
            keycloak.KeyCloakRequestTypeChoices,
            keycloak.KeyCloakRequestMethodChoices.GET,
            keycloak.KeyCloakPanelTypeChoices.ADMIN,
            role_id=role_id
        )
    except Exception as e:
        return JSONResponse({"detail": str(e)},
                            status_code=status.HTTP_400_BAD_REQUEST)

    return JSONResponse(response, status_code=status.HTTP_200_OK)


@router.post('/groups/{group_id}/roles/assign/',tags=['KeyCloak - Admin'])
@require_any_group(*access['assign_role_group'])
def assign_roles_to_group(
        roles: AssignRoleGroupManySchema,
        group_id: str,
        user: UserPayload = Depends(authenticate)
):
    """
    Assigning roles to group
    """
    keycloak = KeyCloakConfidentialClient()

    try:
        response = keycloak.send_request(
            keycloak.KeyCloakRequestTypeChoices.ASSIGN_ROLE_GROUP,
            keycloak.KeyCloakRequestTypeChoices,
            keycloak.KeyCloakRequestMethodChoices.POST,
            keycloak.KeyCloakPanelTypeChoices.ADMIN,
            group_id=group_id,
            roles=roles.dict()
        )

    except Exception as e:
        return JSONResponse({"detail": str(e)}, status_code=status.HTTP_400_BAD_REQUEST)

    return JSONResponse(response, status_code=status.HTTP_200_OK)


@router.post('/users/group/join/',tags=['KeyCloak - Admin'])
@require_any_group(*access['join_user_group'])
def user_join_to_group(
        body: UserJoinGroupSchema,
        user: UserPayload = Depends(authenticate)
):
    """
    User membership to a group using their IDs
    """
    keycloak = KeyCloakConfidentialClient()

    try:
        response = keycloak.send_request(
            keycloak.KeyCloakRequestTypeChoices.USER_JOIN_GROUP,
            keycloak.KeyCloakRequestTypeChoices,
            keycloak.KeyCloakRequestMethodChoices.PUT,
            keycloak.KeyCloakPanelTypeChoices.ADMIN,
            user_id=body.user_id,
            group_id=body.group_id,
        )

    except Exception as e:
        return JSONResponse({"detail": str(e)}, status_code=status.HTTP_400_BAD_REQUEST)

    return JSONResponse(response, status_code=status.HTTP_200_OK)



from fastapi import HTTPException

class SSOUserNotFound(HTTPException):
    def __init__(self, user_id: str):
        super().__init__(
            status_code=400,
            detail={"error": "User not found in SSO", "user_id": user_id}
        )



import redis
import json
from typing import Any, Optional
from fastapi_keycloak_sso.initializer import KeyCloakInitializer

redis_host = KeyCloakInitializer.redis_host


class RedisSSOCache:
    def __init__(self, host=redis_host, port=6379, db=0):
        self.client = redis.Redis(host=host, port=port, db=db, decode_responses=True)

    def get_custom_class_cached_value(self, cache_key: str) -> Optional[Any]:
        data = self.client.get(cache_key)
        if data:
            return json.loads(data)
        return None

    def set_custom_class_cache_value(self, cache_key: str, value: Any, timeout: int = 3600) -> None:
        self.client.set(cache_key, json.dumps(value), ex=timeout)



import gettext
import datetime
import time
import base64
import hashlib
import os
from enum import Enum
from typing import Type, Any, Optional
from urllib.parse import urlencode

import requests
from jose import exceptions as jose_exceptions
from jose import jwt
from fastapi import Response, status, Request
from .initializer import KeyCloakInitializer



gettext.bindtextdomain('myapp', 'locales')
gettext.textdomain('myapp')
_ = gettext.gettext


_jwks = None

class KeyCloakBaseManager(KeyCloakInitializer):
    class KeyCloakException(Exception):
        pass

    class KeyCloakNotFoundException(Exception):
        pass

    class KeyCloakGroupRoleChoices(str, Enum):
        MANAGER = "MANAGER"
        ASSISTANT = "ASSISTANT"
        EMPLOYEE = "EMPLOYEE"

    class KeyCloakClientTypeChoices(str, Enum):
        CONFIDENTIAL = "CONFIDENTIAL"
        PUBLIC = "PUBLIC"

    class KeyCloakPanelTypeChoices(str, Enum):
        ADMIN = "ADMIN"
        USER = "USER"

    class KeyCloakRequestMethodChoices(str, Enum):
        GET = "GET"
        POST = "POST"
        PUT = "PUT"
        DELETE = "DELETE"

    class KeyCloakSaveTokenMethodChoices(str, Enum):
        COOKIE = "COOKIE"
        HEADER = "HEADER"

    def __init__(self, save_token_method: str = KeyCloakSaveTokenMethodChoices.HEADER):
        self.save_token_method =  save_token_method

    def _get_jwks(self):
        global _jwks
        if not _jwks:
            resp = requests.get(
                self.jwks_url,
                verify=False
            )
            resp.raise_for_status()
            _jwks = resp.json()
        return _jwks

    def decode_token(self, token: str):
        jwks = self._get_jwks()
        try:
            decoded_content = jwt.decode(
                token,
                jwks,
                algorithms=[self.algorithms],
                audience=self.user_audience
            )
        except (jose_exceptions.JWTError, jose_exceptions.ExpiredSignatureError, jose_exceptions.JWTClaimsError) as e:
            raise self.KeyCloakException(f"Failed to decode token : {str(e)}")
        return decoded_content

    @staticmethod
    def _generate_code_verifier() -> str:
        return base64.urlsafe_b64encode(os.urandom(32)).rstrip(b'=').decode('utf-8')

    @staticmethod
    def _generate_code_challenge(verifier: str) -> str:
        digest = hashlib.sha256(verifier.encode('utf-8')).digest()
        return base64.urlsafe_b64encode(digest).rstrip(b'=').decode('utf-8')

    @classmethod
    def validate_enum_value(cls, value: str, enum_class: Type[Enum]):
        if value not in [e.value for e in enum_class]:
            raise cls.KeyCloakException(f"Value '{value}' is not valid for {enum_class.__name__}")

    @staticmethod
    def _build_url(base_url: str, endpoint: str, params: dict = None) -> str:
        url = f"{base_url.rstrip('/')}/{endpoint.lstrip('/')}"
        if params:
            url = f"{url}?{urlencode(params)}"
        return url

    def _get_headers(self, extra_headers: dict = None) -> dict:
        headers = {}
        if extra_headers:
            headers.update(extra_headers)
        return headers

    def _get_request_data(
            self,
            *,
            endpoint: str,
            request_method: KeyCloakRequestMethodChoices,
            post_data: Any = None,
            extra_headers: dict = None,
            is_admin: bool = False
        ) -> Any:
            url = f"{self.base_admin_url}{endpoint}" if is_admin else f"{self.base_panel_url}{endpoint}"
            headers = self._get_headers(extra_headers)

            try:
                response = None

                if request_method == self.KeyCloakRequestMethodChoices.GET:
                    response = requests.get(url, headers=headers, verify=False)

                elif request_method == self.KeyCloakRequestMethodChoices.POST:
                    content_type = headers.get("Content-Type", "").lower()
                    if "application/json" in content_type:
                        response = requests.post(url, headers=headers, json=post_data, verify=False)
                    else:
                        response = requests.post(url, headers=headers, data=post_data, verify=False)

                elif request_method == self.KeyCloakRequestMethodChoices.PUT:
                    content_type = headers.get("Content-Type", "").lower()
                    if "application/json" in content_type:
                        response = requests.put(url, headers=headers, json=post_data, verify=False)
                    else:
                        response = requests.put(url, headers=headers, data=post_data, verify=False)

                elif request_method == self.KeyCloakRequestMethodChoices.DELETE:
                    response = requests.delete(url, headers=headers, data=post_data, verify=False)

                if response is not None:
                    response.raise_for_status()
                    if response.status_code in (200, 201, 204):
                        if not response.content or not response.content.strip():
                            return {"detail": "Request successful"}
                        try:
                            return response.json()
                        except ValueError:
                            return {"detail": "Non-JSON response", "raw": response.text}

            except requests.HTTPError as http_err:
                if http_err.response.status_code == 404:
                    raise self.KeyCloakNotFoundException(f"Not found: {url}")
                elif http_err.response.status_code == 409:
                    raise self.KeyCloakException("Conflict: object already exists")
                else:
                    raise self.KeyCloakException(str(http_err))
            except Exception as err:
                raise self.KeyCloakException(str(err))

    def send_request(
        self,
        request_type: Enum,
        request_type_choices: Type[Enum],
        request_method: KeyCloakRequestMethodChoices,
        panel_type: KeyCloakPanelTypeChoices,
        *args,
        **kwargs
    ):
        
        if request_method not in self.KeyCloakRequestMethodChoices:
            raise self.KeyCloakException("Invalid request method")
        if panel_type not in self.KeyCloakPanelTypeChoices:
            raise self.KeyCloakException("Invalid panel type")
        if request_type not in request_type_choices:
            raise self.KeyCloakException("Invalid request type")


        get_data_method = getattr(self, f"_{request_method.lower()}_{request_type.name.lower()}", None)
        if not get_data_method or not callable(get_data_method):
            raise self.KeyCloakException("Data get method for Keycloak is not defined")
        return get_data_method(*args, **kwargs)
    
    @staticmethod
    def get_token_from_header(request):
        auth = request.META.get('HTTP_AUTHORIZATION', '')
        if auth.startswith('Bearer '):
            return auth.split(' ')[1]
        return None

    @staticmethod
    def get_token_from_cookie(request, key: str) -> str:
        token = request.COOKIES.get(key, None)
        return token

    @staticmethod
    def get_token_from_request(request: Request, key: str) -> str | None:
        token = None

        try:
            body = request.json()
            token = body.get(key)
        except Exception:
            pass

        if token is None:
            try:
                form = request.form()
                token = form.get(key)
            except Exception:
                pass

        if token is None:
            token = request.query_params.get(key)
        return token

    def get_token(self, request, key: str = '') -> str:
        token = None
        if self.save_token_method == self.KeyCloakSaveTokenMethodChoices.COOKIE:
            token = self.get_token_from_cookie(request, key)
        elif self.save_token_method == self.KeyCloakSaveTokenMethodChoices.HEADER:
            if key == 'access_token':
                token = self.get_token_from_header(request)
            else:
                token = self.get_token_from_request(request, key)
        return token


class KeyCloakConfidentialClient(KeyCloakBaseManager):
    default_client_roles = [
        "offline_access",
        "uma_authorization",
        "default-roles-markaz",
    ]

    class KeyCloakRequestTypeChoices(str, Enum):
        CLIENT_CREDENTIALS_ACCESS_TOKEN = "CLIENT_CREDENTIALS_ACCESS_TOKEN"
        PASSWORD_ACCESS_TOKEN = "PASSWORD_ACCESS_TOKEN"
        REFRESH_ACCESS_TOKEN = "REFRESH_ACCESS_TOKEN"
        INTROSPECT_TOKEN = "INTROSPECT_TOKEN"
        USER_INFO = "USER_INFO"
        LOGOUT = "LOGOUT"
        JWKS_VERIFY = "JWKS_VERIFY"
        GROUPS = "GROUPS"
        USERS = "USERS"
        USER_ROLES = "USER_ROLES"
        USER_GROUPS = "USER_GROUPS"
        CLIENT_ROLES = "CLIENT_ROLES"
        ASSIGN_ROLE_GROUP = "ASSIGN_ROLE_GROUP"
        USER_JOIN_GROUP = "USER_JOIN_GROUP"
        FIND_GROUP = "FIND_GROUP"

    KEYCLOAK_TOKEN_CACHE_KEY = "keycloak_credentials_client_access_token"
    KEYCLOAK_TOKEN_EXPIRE_KEY = "keycloak_credentials_client_access_token_expiry"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client_type = self.KeyCloakClientTypeChoices.CONFIDENTIAL

    def set_client_access_token(self, headers: dict) -> dict:
        access_token = self.get_cached_access_token()
        headers.update({"Authorization": f"Bearer {access_token}"})
        return headers

    def _get_jwks(self):
        global _jwks
        if not _jwks:
            resp = requests.get(self.jwks_url, verify=False)
            resp.raise_for_status()
            _jwks = resp.json()
        return _jwks
    

    @staticmethod
    def set_httponly_cookie(
        key: str,
        value: str,
        response: Optional[Response] = None,
        expires_hours: int = 1
    ) -> Response:
        if not response:
            response = Response(status_code=status.HTTP_200_OK)
        expire_time = datetime.datetime.utcnow() + datetime.timedelta(hours=expires_hours)
        response.set_cookie(
            key=key,
            value=value,
            httponly=True,
            secure=False,
            expires=expire_time.strftime("%a, %d-%b-%Y %H:%M:%S GMT"),
            samesite="Lax"
        )
        return response
    
    @staticmethod
    def _build_filter_url(*, base_url: str, extra_query_params: dict = None, detail_pk: str | None = None) -> str:
        query_params = {}
        base_url = f'{base_url}/{detail_pk}' if detail_pk else base_url
        if extra_query_params:
            query_params.update(extra_query_params)
        return f"{base_url}?{urlencode(query_params)}" if query_params else base_url

    def get_cached_access_token(self):

        cache_dict = getattr(self, "_token_cache", {})
        access_token = cache_dict.get(self.KEYCLOAK_TOKEN_CACHE_KEY)
        expiry_time = cache_dict.get(self.KEYCLOAK_TOKEN_EXPIRE_KEY)
        if access_token and expiry_time and expiry_time > time.time():
            return access_token

        return self._post_client_credentials_access_token()



    def _post_client_credentials_access_token(self, *args, **kwargs):
        endpoint = "/protocol/openid-connect/token"
        endpoint = self._build_filter_url(base_url=endpoint, **kwargs)

        post_data = {
            "grant_type": "client_credentials",
            "client_id": self.client_id,
            "client_secret": self.client_secret,
        }

        response_data = self._get_request_data(
            endpoint=endpoint,
            request_method=self.KeyCloakRequestMethodChoices.POST,
            extra_headers=None,
            post_data=post_data,
            is_admin=False
        )

        if response_data:
            access_token = response_data.get("access_token")
            expires_in = response_data.get("expires_in", 300)  # default 5 mins

            self._token_cache = getattr(self, "_token_cache", {})
            self._token_cache[self.KEYCLOAK_TOKEN_CACHE_KEY] = access_token
            self._token_cache[self.KEYCLOAK_TOKEN_EXPIRE_KEY] = time.time() + expires_in - 30

            return access_token

        raise self.KeyCloakException("Failed to retrieve client credentials access token")


    def _post_password_access_token(self, username: str, password: str, *args, **kwargs):
        endpoint = "/protocol/openid-connect/token"
        endpoint = self._build_filter_url(base_url=endpoint, **kwargs)
        post_data = {
            "grant_type": "password",
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "username": username,
            "password": password,
        }
        response_data = self._get_request_data(
            endpoint=endpoint,
            request_method=self.KeyCloakRequestMethodChoices.POST,
            extra_headers=None,
            post_data=post_data,
            is_admin=False
        )

        if response_data:
            return response_data.get('access_token', None)
        else:
            raise self.KeyCloakException(_("Failed to retrieve data"))


    def _post_refresh_access_token(self, refresh_token: str, client_id: str = None, *args, **kwargs) -> dict:
        endpoint = "/protocol/openid-connect/token"
        endpoint = self._build_filter_url(base_url=endpoint, **kwargs)
        post_data = {
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
        }
        if client_id:
            post_data.update({
                "client_id": client_id,
            })
        else:
            post_data.update({
                "client_id": self.client_id,
                "client_secret": self.client_secret,
            })
        response_data = self._get_request_data(
            endpoint=endpoint,
            request_method=self.KeyCloakRequestMethodChoices.POST,
            extra_headers=None,
            post_data=post_data,
            is_admin=False
        )

        if not response_data:
            raise self.KeyCloakException(_("Failed to retrieve data"))
        return response_data


    def _get_groups(self , *args, **kwargs) -> dict:
        """
        Retrieves all groups from Keycloak using the Admin REST API.
        Requires a valid admin-level access token.
        """
        endpoint = "/groups"
        endpoint = self._build_filter_url(base_url=endpoint, **kwargs)
        extra_headers = {
            "Content-Type": "application/json"
        }
        extra_headers = self.set_client_access_token(extra_headers)
        response_data = self._get_request_data(
            endpoint=endpoint,
            request_method=self.KeyCloakRequestMethodChoices.GET,
            extra_headers=extra_headers,
            post_data=None,
            is_admin=True
        )

        if not response_data:
            raise self.KeyCloakException(_("Failed to retrieve groups from Keycloak"))

        return response_data

    def _get_users(self, *args, **kwargs) -> dict:
        """
        Retrieves all users from Keycloak using the Admin REST API.
        Requires a valid admin-level access token.
        """
        endpoint = "/users"
        endpoint = self._build_filter_url(base_url=endpoint, **kwargs)
        extra_headers = {
            "Content-Type": "application/json"
        }
        extra_headers = self.set_client_access_token(extra_headers)
        response_data = self._get_request_data(
            endpoint=endpoint,
            request_method=self.KeyCloakRequestMethodChoices.GET,
            extra_headers=extra_headers,
            post_data=None,
            is_admin=True
        )

        if not response_data:
            raise self.KeyCloakException(_("Failed to retrieve users from Keycloak"))

        return response_data


    # for create group
    def _post_groups(self , name: str , group_parent_id: str = None):
        endpoint = '/groups'

        if group_parent_id:
            endpoint = f'/groups/{group_parent_id}/children'

        endpoint = self._build_filter_url(base_url=endpoint)
        extra_headers = {
            "Content-Type": "application/json"
        }
        data = {
            'name': name
        }
        extra_headers = self.set_client_access_token(extra_headers)
        response_data = self._get_request_data(
            endpoint=endpoint,
            request_method=self.KeyCloakRequestMethodChoices.POST,
            extra_headers=extra_headers,
            post_data=data,
            is_admin=True
        )
        return response_data

    def _delete_groups(self , group_id: str):
        endpoint = f'/groups/{group_id}'
        endpoint = self._build_filter_url(base_url=endpoint)
        extra_headers = {
            "Content-Type": "application/json"
        }
        extra_headers = self.set_client_access_token(extra_headers)
        response_data = self._get_request_data(
            endpoint=endpoint,
            request_method=self.KeyCloakRequestMethodChoices.DELETE,
            extra_headers=extra_headers,
            post_data=None,
            is_admin=True
        )
        return response_data

    def _get_client_roles(self , role_id: str = None):
        endpoint = f'/clients/{self.client_pk}/roles'
        if role_id:
            endpoint = f'/roles-by-id/{role_id}'

        endpoint = self._build_filter_url(base_url=endpoint)
        extra_headers = {
            "Content-Type": "application/json"
        }
        extra_headers = self.set_client_access_token(extra_headers)
        response_data = self._get_request_data(
            endpoint=endpoint,
            request_method=self.KeyCloakRequestMethodChoices.GET,
            extra_headers=extra_headers,
            post_data=None,
            is_admin=True
        )
        return response_data

    def _post_assign_role_group(self, group_id: str , roles: dict):
        endpoint = f'/groups/{group_id}/role-mappings/clients/{self.client_pk}'
        endpoint = self._build_filter_url(base_url=endpoint)
        extra_headers = {
            "Content-Type": "application/json"
        }
        extra_headers = self.set_client_access_token(extra_headers)

        role_objects = roles['roles']

        data = []
        for role in role_objects:
            role_id = role['role_id']
            role_name = role['role_name']
            data.append({
                'id' : role_id,
                'name' : role_name
            })

        response_data = self._get_request_data(
            endpoint=endpoint,
            request_method=self.KeyCloakRequestMethodChoices.POST,
            extra_headers=extra_headers,
            post_data=data,
            is_admin=True
        )
        return response_data

    def _put_user_join_group(self,user_id,group_id):
        endpoint = f'/users/{user_id}/groups/{group_id}'
        endpoint = self._build_filter_url(base_url=endpoint)
        extra_headers = {
            "Content-Type": "application/json"
        }
        extra_headers = self.set_client_access_token(extra_headers)

        response_data = self._get_request_data(
            endpoint=endpoint,
            request_method=self.KeyCloakRequestMethodChoices.PUT,
            extra_headers=extra_headers,
            post_data=None,
            is_admin=True
        )
        return response_data



import os
from enum import Enum
from typing import Type
from urllib.parse import urlencode
import logging
from .caching import RedisSSOCache
from fastapi_keycloak_sso.keycloak import KeyCloakConfidentialClient
from fastapi_keycloak_sso.keycloak import KeyCloakInitializer

logger = logging.getLogger(__name__)

redis_host = KeyCloakInitializer.redis_host

sso_cache = RedisSSOCache(host=redis_host, port=6379, db=0)


class SSOClass:
    class SSOClassException(Exception):
        pass

    class SSOClassNotFoundException(Exception):
        pass

    class CompanyGroupRoleChoices(Enum):
        MANAGER = "MANAGER"
        ASSISTANT = "ASSISTANT"
        EMPLOYEE = "EMPLOYEE"

    class SSODataTypeChoices(Enum):
        USER = "USER"
        USER_ROLE = "USER_ROLE"
        COMPANY_GROUP = "COMPANY_GROUP"

    class SSODataFormChoices(Enum):
        DETAIL = "DETAIL"
        LIST = "LIST"
        CUSTOM = "CUSTOM"

    class SSOFieldTypeChoices(Enum):
        GROUP = "GROUP"
        USER = "USER"
        ROLE = "ROLE"

    sso_request_exceptions = (
        SSOClassException,
        SSOClassNotFoundException,
        KeyCloakConfidentialClient.KeyCloakException,
        KeyCloakConfidentialClient.KeyCloakNotFoundException,
    )

    def __init__(self):
        self.keycloak = KeyCloakConfidentialClient()

    @classmethod
    def validate_enums_value(cls, value: str, enums_class: Type[Enum]):
        if value not in [e.value for e in enums_class]:
            raise cls.SSOClassException("Value is not exists in that enums")

    @staticmethod
    def _build_filter_url(*, base_url: str, ids_filtering_list: list = None, id_range_filtering_list: list = None):
        query_params = {}
        return f"{base_url}?{urlencode(query_params)}" if query_params else base_url

    def get_sso_data(self, data_type: Enum, data_form: Enum, *args, **kwargs):
        self.validate_enums_value(data_type.value, data_type.__class__)
        self.validate_enums_value(data_form.value, data_form.__class__)
        if data_form == self.SSODataFormChoices.DETAIL and 'pk' not in kwargs:
            raise self.SSOClassException("Get detail of a object need object pk")
        get_data_method = getattr(self, f'get_{data_type.value.lower()}_{data_form.value.lower()}_data', None)
        if not get_data_method:
            raise self.SSOClassException("Data get method for sso data is not valid")
        return get_data_method(*args, **kwargs)

    def get_user_detail_data(self, pk, *args, **kwargs):
        cache_key = f"user_detail_{pk}"
        data = sso_cache.get_custom_class_cached_value(cache_key)
        if data:
            logger.info(f"Cache HIT for user {pk}")
            return data

        logger.info(f"Cache MISS for user {pk}")
        user_data = self.keycloak.send_request(
            self.keycloak.KeyCloakRequestTypeChoices.USERS,
            self.keycloak.KeyCloakRequestTypeChoices,
            self.keycloak.KeyCloakRequestMethodChoices.GET,
            self.keycloak.KeyCloakPanelTypeChoices.ADMIN,
            detail_pk=pk,
        )
        if user_data:
            sso_cache.set_custom_class_cache_value(cache_key, user_data, timeout=3600)
            return user_data
        raise self.SSOClassException("Failed to retrieve data with user ID")

    def get_user_list_data(self, *args, **kwargs):
        cache_key = "user_list"
        data = sso_cache.get_custom_class_cached_value(cache_key)
        if data:
            logger.info("Cache HIT for user list")
            return data

        logger.info("Cache MISS for user list")
        users_data = self.keycloak.send_request(
            self.keycloak.KeyCloakRequestTypeChoices.USERS,
            self.keycloak.KeyCloakRequestTypeChoices,
            self.keycloak.KeyCloakRequestMethodChoices.GET,
            self.keycloak.KeyCloakPanelTypeChoices.ADMIN,
        )
        if users_data:
            sso_cache.set_custom_class_cache_value(cache_key, users_data, timeout=3600)
            return users_data
        raise self.SSOClassException("Failed to retrieve user list data")

    def get_company_group_detail_data(self, pk, *args, **kwargs):
        cache_key = f"company_group_detail_{pk}"
        data = sso_cache.get_custom_class_cached_value(cache_key)
        if data:
            logger.info(f"Cache HIT for company group {pk}")
            return data

        logger.info(f"Cache MISS for company group {pk}")
        group_data = self.keycloak.send_request(
            self.keycloak.KeyCloakRequestTypeChoices.GROUPS,
            self.keycloak.KeyCloakRequestTypeChoices,
            self.keycloak.KeyCloakRequestMethodChoices.GET,
            self.keycloak.KeyCloakPanelTypeChoices.ADMIN,
            detail_pk=pk
        )
        if group_data:
            sso_cache.set_custom_class_cache_value(cache_key, group_data, timeout=3600)
            return group_data
        raise self.SSOClassException("Failed to retrieve company groups detail data")



from fastapi import Depends
from fastapi_keycloak_sso.initializer import KeyCloakInitializer
from fastapi_keycloak_sso.auth import UserPayload, authenticate

def check_user_permission_access(
        user,
        role_titles: list[str],
        group_titles: list[str],
        group_roles: list[str],
        match_group_roles: bool = False,
        permissive: bool = False,
) -> bool :
    """
    Check permissions for either CustomUser or UserPayload.
    Extracts roles and groups from payload if needed.
    """

    # Realm roles
    user_roles = [r.lower() for r in getattr(user, "realm_access", {}).get("roles") or []]

    # Client roles
    client_name = KeyCloakInitializer.client_name
    user_client_roles = [
        r.lower() for r in (getattr(user, "resource_access", {}).get(client_name, {}).get("roles") or [])
    ]

    # Groups
    user_groups_raw = getattr(user, "groups", []) or []

    # Normalize input
    role_titles = [r.lower() for r in role_titles]
    group_titles = [g.lower() for g in group_titles]
    group_roles = [r.lower() for r in group_roles]

    # Parse groups (from paths like '/group_1/managers')
    parsed_user_groups = []
    for group_path in user_groups_raw:
        parts = group_path.strip("/").split("/")
        for part in parts:
            parsed_user_groups.append((part.lower(), None))

    # Rule 1: required roles
    denied_required_role = False
    require_required_role = False
    for required_role in role_titles:
        require_required_role = True
        if required_role not in user_roles and required_role not in user_client_roles:
            denied_required_role = True

    # Rule 2: required group titles
    denied_group_title_role = False
    require_group_title_role = False
    user_group_names = [group for group, _ in parsed_user_groups]
    for required_group in group_titles:
        require_group_title_role = True
        if required_group not in user_group_names:
            denied_group_title_role = True

    # Rule 3: required group roles
    denied_group_role = False
    require_group_role = False
    if group_roles:
        require_group_role = True
        matched = False
        for group, role in parsed_user_groups:
            if match_group_roles:
                if group in group_titles and role in group_roles:
                    matched = True
                    break
            else:
                if role in group_roles:
                    matched = True
                    break
        if not matched:
            denied_group_role = True

    # Adjust denied flags
    denied_required_role = denied_required_role if require_required_role else False
    denied_group_title_role = denied_group_title_role if require_group_title_role else False
    denied_group_role = denied_group_role if require_group_role else False

    # Return final decision
    if permissive:
        return not (denied_required_role and denied_group_title_role and denied_group_role)
    else:
        return not (denied_required_role or denied_group_title_role or denied_group_role)



import inspect
from functools import wraps
from fastapi.responses import JSONResponse
from fastapi import HTTPException, status, Depends
from .sso.utils import check_user_permission_access
from .auth import UserPayload,authenticate


def _is_coroutine(func):
    return callable(getattr(func, "__await__", None))

def check_permission_decorator(
    role_titles=None,
    group_titles=None,
    group_roles=None,
    match_group_roles=False,
    permissive=False,
):
    role_titles = role_titles or []
    group_titles = group_titles or []
    group_roles = group_roles or []

    def decorator(func):
        @wraps(func)
        async def wrapper(*args, user: UserPayload = Depends(authenticate), **kwargs):
            if not user or not getattr(user, "id", None):
                raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail="Authentication required")

            if not role_titles and not group_titles and not group_roles:
                return await func(*args, user=user, **kwargs) if _is_coroutine(func) else func(*args, user=user, **kwargs)

            has_access = check_user_permission_access(
                user=user,
                role_titles=role_titles,
                group_titles=group_titles,
                group_roles=group_roles,
                match_group_roles=match_group_roles,
                permissive=permissive,
            )

            if not has_access:
                raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail="You are not allowed to access this API")

            return await func(*args, user=user, **kwargs) if _is_coroutine(func) else func(*args, user=user, **kwargs)

        return wrapper
    return decorator


# ---------------- Helper decorators ----------------

def require_roles(*role_titles):
    return check_permission_decorator(role_titles=list(role_titles))


def require_groups(*group_titles):
    return check_permission_decorator(group_titles=list(group_titles))


def require_group_roles(*group_roles):
    return check_permission_decorator(group_roles=list(group_roles))


def require_any_group(*group_titles):
    group_titles = [r.lower() for r in group_titles if r]

    def decorator(func):
        @wraps(func)
        async def wrapper(*args, user: UserPayload = Depends(authenticate), **kwargs):
            if not user or not getattr(user, "id", None):
                raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail="Authentication required")

            if not group_titles:
                return await func(*args, user=user, **kwargs) if _is_coroutine(func) else func(*args, user=user, **kwargs)

            user_groups = getattr(user, 'groups', [])
            if not any(group in [g.lower() for g in user_groups] for group in group_titles):
                raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail="You are not allowed to access this API")

            return await func(*args, user=user, **kwargs) if _is_coroutine(func) else func(*args, user=user, **kwargs)

        return wrapper
    return decorator

def require_any_role(*role_titles):
    role_titles = [r for r in role_titles if r]

    def decorator(func):
        @wraps(func)
        async def wrapper(*args, user: UserPayload = Depends(authenticate), **kwargs):
            if not user or not getattr(user, "id", None):
                raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail="Authentication required")

            if not role_titles:
                return await func(*args, user=user, **kwargs) if _is_coroutine(func) else func(*args, user=user, **kwargs)

            user_roles = set(user.roles.get("client_roles", []) + user.roles.get("realm_roles", []))
            user_roles = [r.lower() for r in user_roles if r]

            if any(role.lower() in user_roles for role in role_titles):
                return await func(*args, user=user, **kwargs) if _is_coroutine(func) else func(*args, user=user, **kwargs)

            raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail="You are not allowed to access this API")

        return wrapper
    return decorator

def require_all_permissions(
    *,
    role_titles=None,
    group_titles=None,
    group_roles=None,
    match_group_roles=False,
    permissive=False
):
    return check_permission_decorator(
        role_titles=role_titles or [],
        group_titles=group_titles or [],
        group_roles=group_roles or [],
        match_group_roles=match_group_roles,
        permissive=permissive
    )



import os
from typing import Any
from .caching import RedisSSOCache
from fastapi_keycloak_sso.keycloak import KeyCloakConfidentialClient
from fastapi_keycloak_sso.initializer import KeyCloakInitializer

redis_host = KeyCloakInitializer.redis_host

sso_cache = RedisSSOCache(host=redis_host, port=6379, db=0)

class CustomGetterObjectClass:
    def __init__(self, payload: dict):
        self.is_exists = bool(payload)
        self._payload = payload
        self.keycloak_klass = KeyCloakConfidentialClient()

    def __getattr__(self, name):
        if not self.is_exists:
            return None
        if name in self._payload:
            return self._payload[name]
        return super().__getattribute__(name)

    def __bool__(self):
        return self.is_exists

    def __repr__(self):
        return f"<CustomGetterObjectClass()>"

    @property
    def pk(self):
        if 'id' in self._payload:
            return str(self._payload['id'])
        elif 'sub' in self._payload:
            return str(self._payload['sub'])
        return None

    def _get_cache_key(self, cache_base_key: str):
        return f"{cache_base_key}_{self.pk}"

    def _get_cached_value(self, cache_base_key: str) -> Any:
        cache_key = self._get_cache_key(cache_base_key)
        return sso_cache.get_custom_class_cached_value(cache_key)

    def _set_cache_value(self, cache_base_key: str, value: Any, timeout: int = 3600) -> None:
        cache_key = self._get_cache_key(cache_base_key)
        sso_cache.set_custom_class_cache_value(cache_key, value, timeout)



from typing import Any, Optional, Type
from sqlalchemy.types import TypeDecorator, String
from .helpers import sso_cache
from .authenticate import CustomUser
from fastapi_keycloak_sso.sso.sso import SSOClass



class CustomSSORelatedField:
    """
    Get field information from cache and if not, get it from keycloak and store it in cache
    """
    def __init__(self, max_length: int = 36):
        self.max_length = max_length

    def _get_sso_field_value(
        self,
        value: str | int,
        sso_method: str,
        cache_key: Optional[str] = None,
        getter_klass: Optional[Type] = None
    ) -> Any:

        key = f"{cache_key}_{value}" if cache_key else f"{str(self.__class__.__name__).lower()}_{value}"
        data = sso_cache.get_custom_class_cached_value(key)
        if not data:
            sso_client = SSOClass()
            if not hasattr(sso_client, sso_method):
                raise Exception("SSO Class hasn't specified method")
            try:
                data = getattr(sso_client, sso_method)(pk=value)
                sso_cache.set_custom_class_cache_value(key, data, timeout=3600)
            except Exception:
                data = None

        getter_klass = getter_klass or CustomUser
        return getter_klass(payload=data)


class SSOUserField(TypeDecorator, CustomSSORelatedField):
    """
    Type the field for the desired id value in the database, preferably sqlalchemy
    """
    impl = String
    cache_base_key = "sso_user"
    cache_ok = True

    def __init__(self, *args, **kwargs):
        TypeDecorator.__init__(self, *args, **kwargs)
        CustomSSORelatedField.__init__(self)

    def process_result_value(self, value: Any, dialect) -> Optional[CustomUser]:
        if value is None:
            return None
        return CustomUser(False, {"id": value})

    def process_bind_param(self, value: Any, dialect) -> Optional[str]:
        from .schemas import LazySSOUser

        if value is None:
            return None

        if isinstance(value, CustomUser):
            return value.pk

        if isinstance(value, str):
            return value

        if isinstance(value, LazySSOUser):
            return str(value)

        raise ValueError("SSOUserField only accepts CustomUser, LazySSOUser, or string user_id.")

    def get_full_data(self, value: str | int) -> Any:

        key = f"{self.cache_base_key}_{value}"
        return self._get_sso_field_value(
            value=value,
            sso_method="get_user_detail_data",
            cache_key=key,
            getter_klass=CustomUser
        )



from .initializer import KeyCloakInitializer
from fastapi import HTTPException, status


class GroupAccess(KeyCloakInitializer):
    """
    You can check if a user has the allowed group by sending the name of the desired group and
    user and setting the admin group in .env.
    """

    def require_all_groups(self, user , groups_name: list):
        groups_list = user.groups
        admin_groups = self.admin_groups.split(',')

        if (not any(admin in groups_list for admin in admin_groups) and
            not all(item in groups_list for item in groups_name)):
            raise HTTPException(
                detail='You are not allowed to access this API',
                status_code=status.HTTP_403_FORBIDDEN
            )
        return True

    def require_any_groups(self, user, groups_names: list):
        groups_list = user.groups
        admin_groups = self.admin_groups.split(',')

        if not (
                any(admin in groups_list for admin in admin_groups)
                or any(item in groups_list for item in groups_names)
        ):
            raise HTTPException(
                detail='You are not allowed to access this API',
                status_code=status.HTTP_403_FORBIDDEN
            )
        return True



import os
from dotenv import load_dotenv, find_dotenv


env_path = find_dotenv(usecwd=True)
load_dotenv(env_path)


class KeyCloakInitializer:
    realm = os.getenv('KEYCLOAK_REALM')
    client_id = os.getenv('KEYCLOAK_CLIENT_ID')
    client_pk = os.getenv('KEYCLOAK_CLIENT_PK')
    client_title = os.getenv('KEYCLOAK_CLIENT_TITLE')
    client_name = os.getenv('KEYCLOAK_CLIENT_NAME')
    algorithms = os.getenv('KEYCLOAK_ALGORITHMS')
    secret_key_algorithm = os.getenv('KEYCLOAK_SECRET_KEY_ALGORITHM')
    user_audience = "account"
    base_prefix_url = os.getenv('KEYCLOAK_SERVER_URL')
    base_panel_url = f'{base_prefix_url}/realms/{realm}'
    base_admin_url = f'{base_prefix_url}/admin/realms/{realm}'
    jwks_url = f"{base_panel_url}/protocol/openid-connect/certs"
    issuer_prefix = os.getenv('KEYCLOAK_ISSUER_PREFIX')
    issuer = f'{issuer_prefix}/realms/{realm}'
    client_secret = os.getenv('KEYCLOAK_CLIENT_SECRET')
    admin_groups = os.getenv('ADMIN_GROUPS')  # list set in env split(',')
    service_base_url = os.getenv('SSO_SERVICE_BASE_URL')
    redis_host = os.getenv('KEYCLOAK_REDIS_HOST')


class RoutAccessInitializer:
    group_read_access = (os.getenv('KEYCLOAK_GROUP_READ_ACCESS') or "").split(',') or None
    group_find_access = (os.getenv('KEYCLOAK_GROUP_FIND_ACCESS') or "").split(',') or None
    group_create_access = (os.getenv('KEYCLOAK_GROUP_CREATE_ACCESS') or "").split(',') or None
    group_delete_access = (os.getenv('KEYCLOAK_GROUP_DELETE_ACCESS') or "").split(',') or None
    user_read_access = (os.getenv('KEYCLOAK_USER_READ_ACCESS') or "").split(',') or None
    role_read_access = (os.getenv('KEYCLOAK_ROLE_READ_ACCESS') or "").split(',') or None
    assign_role_group_access = (os.getenv('KEYCLOAK_ASSIGN_ROLE_GROUP_ACCESS') or "").split(',') or None
    join_user_group_access = (os.getenv('KEYCLOAK_JOIN_USER_GROUP_ACCESS') or "").split(',') or None



from jose import jwt, JWSError, ExpiredSignatureError
from fastapi import Depends, HTTPException, status
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from .initializer import KeyCloakInitializer


ALGORITHM = KeyCloakInitializer.algorithms
SECRET_KEY_ALGORITHM = KeyCloakInitializer.secret_key_algorithm


authentication = HTTPBearer()

class UserPayload:
    def __init__(self, payload_dict: dict):
        self._payload = payload_dict
        self.__dict__.update(payload_dict)

    @property
    def id(self):
        return self._payload.get("sub")

    @property
    def username(self):
        return self._payload.get("preferred_username")

    @property
    def first_name(self):
        return self._payload.get("given_name")
    
    @property
    def last_name(self):
        return self._payload.get("family_name")
    
    @property
    def full_name(self):
        return self._payload.get("name")

    @property
    def groups(self):
        return self._payload.get("groups", [])

    @property
    def email(self):
        return self._payload.get("email")

    @property
    def roles(self):
        resource_access = self._payload.get("resource_access", {}) or {}
        realm_access = self._payload.get("realm_access", {}) or {}

        client_roles = (
            resource_access.get(KeyCloakInitializer.client_name, {})
                           .get("roles", [])
        )

        realm_roles = realm_access.get("roles", [])

        return {
            "client_roles": client_roles,
            "realm_roles": realm_roles
        }
    
    @property
    def profile(self):
        return {
            'id': self.id,
            'username': self.username,
            'first_name': self.first_name,
            'last_name': self.last_name,
            'full_name': self.full_name,
            'roles': self.roles,
            'groups': self.groups
        }

    @property
    def companies(self):
        companies = self._payload.get("companies", [])
        return companies


async def authenticate(credentials: HTTPAuthorizationCredentials = Depends(authentication)):
    token = credentials.credentials

    auth_error = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail='Invalid authentication credentials',
        headers={'WWW-Authenticate': 'Bearer'}
    )

    if not ALGORITHM or not SECRET_KEY_ALGORITHM:
        raise HTTPException(
            status_code=400,
            detail="ERROR: Keycloak ALGORITHM or SECRET_KEY_ALGORITHM is not set in environment or initializer.")

    try:

        if ALGORITHM.upper() == "RS256":
            key = f"-----BEGIN PUBLIC KEY-----\n{SECRET_KEY_ALGORITHM}\n-----END PUBLIC KEY-----"
        else:
            key = SECRET_KEY_ALGORITHM

        payload_dict = jwt.decode(token, key, algorithms=[ALGORITHM], options={"verify_aud": False})
        payload = UserPayload(payload_dict)

        if payload.id is None:
            raise auth_error

    except ExpiredSignatureError:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail='Token signature has expired.'
        )
    except JWSError:
        raise auth_error

    return payload



from typing import Any, get_origin, get_args, Optional, Union
from pydantic_core import core_schema
from fastapi_keycloak_sso.sso.authenticate import CustomUser
from fastapi_keycloak_sso.sso.fields import SSOUserField
from .exceptions import SSOUserNotFound



class LazySSOUser:
    def __init__(self, pk: str, sso_field: SSOUserField):
        self.pk = pk
        self._sso_field = sso_field

    def get_full_data(self):
        data: CustomUser = self._sso_field.get_full_data(self.pk)
        if hasattr(data, "to_dict"):
            return data.to_dict()
        return dict(data)

    def __str__(self):
        return self.pk


class SSOUserPydanticField:
    """
    Serializer of user ID field in pydantic
    """

    sso_field: SSOUserField = SSOUserField()

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type, handler):

        def _validate(v: Any, info: Any):

            is_optional = (
                get_origin(source_type) is Optional or
                (get_origin(source_type) is Union and type(None) in get_args(source_type))
            )

            if v is None:
                if is_optional:
                    return None
                raise TypeError("User ID cannot be None")

            if isinstance(v, LazySSOUser):
                user = v
            elif isinstance(v, CustomUser):
                user = LazySSOUser(v.pk, cls.sso_field)
            elif isinstance(v, str):
                user = LazySSOUser(v, cls.sso_field)
            elif hasattr(v, "id"):
                user = LazySSOUser(v.id, cls.sso_field)
            else:
                raise TypeError(f"Invalid user id: {v}")

            return user

        return core_schema.no_info_wrap_validator_function(
            _validate,
            core_schema.any_schema(),
            serialization=core_schema.to_string_ser_schema()
        )


class SSOUserPydanticWithValidation(SSOUserPydanticField):
    """
    Serializer of user ID field in pydantic by checking if the user exists in the system
    """
    @classmethod
    def __get_pydantic_core_schema__(cls, source_type, handler):

        def _validate(v: Any, info: Any):

            is_optional = (
                get_origin(source_type) is Optional or
                (get_origin(source_type) is Union and type(None) in get_args(source_type))
            )

            if v is None:
                if is_optional:
                    return None
                raise TypeError("User ID cannot be None")

            if isinstance(v, LazySSOUser):
                user = v
            elif isinstance(v, CustomUser):
                user = LazySSOUser(v.pk, cls.sso_field)
            elif isinstance(v, str):
                user = LazySSOUser(v, cls.sso_field)
            elif hasattr(v, "id"):
                user = LazySSOUser(v.id, cls.sso_field)
            else:
                raise TypeError(f"Invalid user id: {v}")
            return user

        def validator_with_context(v: Any, info: Any):
            user = _validate(v, info)
            data = user.get_full_data()
            if data is None:
                raise SSOUserNotFound(user_id=str(user))
            return user

        return core_schema.no_info_wrap_validator_function(
            validator_with_context,
            core_schema.any_schema(),
            serialization=core_schema.to_string_ser_schema()
        )

SSOUserPydanticField.sso_field = SSOUserField()



from .helpers import CustomGetterObjectClass
from fastapi_keycloak_sso.keycloak import KeyCloakConfidentialClient


class CustomGroup(CustomGetterObjectClass):
    def __repr__(self):
        return f"<CustomGroup(id={self.id if bool(self) else 'None'})>"

    def get(self, key, default=None):
        """Like dict.get() for payload."""
        return self._payload.get(key, default)

    @property
    def get_user_ids(self):
        """
        Extract a list of unique user IDs from the given data.
        """
        user_ids = set()

        user_groups = self._payload.get("user_groups", [])

        for user_group in user_groups:
            user = user_group.get("user")
            if user and "id" in user:
                user_ids.add(user["id"])

        return list(user_ids)

    @property
    def payload(self) -> dict:
        return self._payload

    @property
    def is_exists(self):
        """Equivalent to DRF logic used in your serializer."""
        return bool(self._payload)



class CustomUser(CustomGetterObjectClass):
    client_title = KeyCloakConfidentialClient.client_title

    def __init__(self, is_authenticated: bool = False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.is_authenticated = is_authenticated

    def __repr__(self):
        return f"<CustomUser(id={self.id if bool(self) else 'None'}, is_authenticated={self.is_authenticated})>"

    def get(self, key, default=None):
        return self._payload.get(key, default)

    # @property
    # def primary_group(self):
    #     for entry in self.groups:
    #         if entry.get('is_main_group'):
    #             return CustomGroup(entry['group'])
    #     return None

    @property
    def groups(self):
        groups_list = []
        if 'groups' not in self.payload:
            return groups_list
        for entry in self.payload['groups']:
            groups_list.append(entry)
        return groups_list

    @property
    def groups_dict_list(self):
        group_list = []
        if 'groups' not in self.payload:
            return group_list
        for path in self.groups:
            parts = path.strip("/").split("/")
            if len(parts) == 2:
                group_name, role_name = parts
                role_name = role_name[:-1] if role_name.endswith('s') else role_name
                group_list.append({
                    'title': group_name,
                    'role': role_name,
                })
        return group_list

    @property
    def groups_parent(self) -> list[str]:
        group_names = []
        if 'groups' not in self.payload:
            return group_names
        for path in self.groups:
            parts = path.strip("/").split("/")
            if len(parts) == 2:
                group_name, _ = parts
                group_names.append(group_name)
        return group_names

    @property
    def group_roles(self) -> list[str]:
        group_roles = []
        if 'groups' not in self.payload:
            return group_roles
        for path in self.groups:
            parts = path.strip("/").split("/")
            if len(parts) == 2:
                _, role_name = parts
                role_name = role_name[:-1] if role_name.endswith('s') else role_name
                group_roles.append(role_name)
        return group_roles

    @property
    def realm_roles(self):
        realm_roles = []
        if 'realm_access' not in self.payload:
            return realm_roles
        if 'roles' not in self.realm_access:
            return realm_roles
        for entry in self.realm_access['roles']:
            if entry in KeyCloakConfidentialClient.default_client_roles or entry.startswith(
                    f'{KeyCloakConfidentialClient.client_name}.'):
                continue
            realm_roles.append(entry)
        return realm_roles

    @property
    def client_roles(self):
        client_roles = []
        if 'resource_access' not in self.payload:
            return client_roles
        for entry in self.resource_access.get(self.client_title, {}).get('roles', []):
            client_roles.append(entry)
        return client_roles

    @property
    def roles(self):
        roles = []
        roles += self.realm_roles
        roles += self.client_roles
        return roles

    @property
    def id(self):
        return self.pk

    @property
    def username(self):
        if 'username' in self.payload:
            return self.payload['username']
        elif 'preferred_username' in self.payload:
            return self.preferred_username
        return str(self.id)

    @property
    def first_name(self):
        if 'first_name' in self.payload:
            return self.payload['first_name']
        elif 'firstName' in self.payload:
            return self.firstName
        elif 'given_name' in self.payload:
            return self.given_name
        return ""

    @property
    def last_name(self):
        if 'last_name' in self.payload:
            return self.payload['last_name']
        elif 'lastName' in self.payload:
            return self.lastName
        elif 'family_name' in self.payload:
            return self.family_name
        return ""

    @property
    def full_name(self):
        if 'full_name' in self.payload and self.payload['full_name']:
            return self.payload['full_name']
        elif 'name' in self.payload and self.payload['name']:
            return self.name
        else:
            first_name = self.first_name
            last_name = self.last_name
            if first_name and last_name:
                return f"{first_name} {last_name}"
            elif first_name:
                return f"{first_name}"
            elif last_name:
                return f"{last_name}"
            else:
                return self.username
        return ""

    @property
    def payload(self) -> dict:
        return self._payload

    @property
    def groups_id(self):
        cached_data = self._get_cached_value('groups_id')

        if cached_data is not None:
            return cached_data

        user_group_data = self.keycloak_klass.send_request(
            self.keycloak_klass.KeyCloakRequestTypeChoices.USER_GROUPS,
            self.keycloak_klass.KeyCloakRequestTypeChoices,
            self.keycloak_klass.KeyCloakRequestMethodChoices.GET,
            self.keycloak_klass.KeyCloakPanelTypeChoices.ADMIN,
            detail_pk=self.id,
        )
        groups_id_list = []
        for entry in user_group_data:
            # groups_id_list.append(entry['group']['id'])
            groups_id_list.append(entry['parentId'])
        self._set_cache_value('groups_id', groups_id_list, 3600)
        return groups_id_list

    def to_dict(self):
        return self._payload


from typing import Optional, List, Any, Dict
from pydantic import BaseModel, field_validator, model_validator
from ..sso.authenticate import CustomGroup


class GiveTokenSchema(BaseModel):
    username: str
    password: str


class LoginSchema(BaseModel):
    token: str
    refresh_token: str
    client_id: str


class GroupSchema(BaseModel):
    id: Optional[str] = None
    title: Optional[str] = None
    subGroups: Optional[List[Dict[str, Any]]] = None

    @model_validator(mode="before")
    def compute_title(cls, values):
        obj = values

        if hasattr(obj, "name"):
            values["title"] = obj.name

        elif isinstance(obj, dict) and obj.get("name"):
            values["title"] = obj["name"]

        return values

    @classmethod
    def from_instance(cls, instance):

        if not isinstance(instance, CustomGroup):
            instance = CustomGroup(payload=instance)

        if not instance.is_exists:
            return cls()  # empty model

        # Convert CustomGroup attributes to dict
        data = {
            "id": getattr(instance, "id", None),
            "name": getattr(instance, "name", None),
            "subGroups": getattr(instance, "subGroups", None),
        }

        return cls(**data)


class UserSchema(BaseModel):
    id: str
    username: str | None
    first_name: str | None
    last_name: str | None
    full_name: str | None
    roles: dict
    groups: list[str]


class PaginatedGroups(BaseModel):
    count: int
    limit: int
    offset: int
    results: List[GroupSchema]


class GroupCreateSchema(BaseModel):
    name: str

class AssignRoleGroupSchema(BaseModel):
    """assigning role to group one object"""
    role_id: str
    role_name: str


class AssignRoleGroupManySchema(BaseModel):
    """assigning role to group many object"""
    roles: List[AssignRoleGroupSchema]


class UserJoinGroupSchema(BaseModel):
    group_id: str
    user_id: str
